/* $Id: mod_midgard.c,v 1.87 2001/03/30 08:18:34 emile Exp $
 *
 * mod_midgard.c  Midgard address translation and user authentication module
 *
 * Copyright (C) 1999 Jukka Zitting <jukka.zitting@iki.fi>
 * Copyright (C) 2000 The Midgard Project ry
 * Copyright (C) 2000 Emile Heyns <emile@iris-advies.com> and Aurora SA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <midgard/apache.h>
#include <midgard/select_db.h>
#include "http_core.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_request.h"
#include <glib.h>

/* This is the module definition part for static Apache modules.
 * It tells the Apache configure script to link to the Midgard library
 * that is required by mod_midgard. If libmidgard is not within a normal
 * library directory (/lib, /usr/lib, /usr/local/lib) then you might want
 * to add -L/path/to/lib to the LDFLAGS variable below.
 *
MODULE-DEFINITION-START
Name: midgard
ConfigStart
  LDFLAGS="$LDFLAGS"
  LIBS=`glib-config --libs `" -lmidgard $LIBS"
  CFLAGS=`glib-config --cflags`
ConfigEnd
MODULE-DEFINITION-END
 */

#ifdef HAVE_MANDRAKE
#define LIBEXEC_DIR "modules"
#else
#define LIBEXEC_DIR "libexec"
#endif

int midgard_authenticate(midgard_request_config * rcfg);
int mgd_get_basic_auth_pw(request_rec * r, const char **usr, const char **pw);

void mgd_note_basic_auth_failure(request_rec *r, midgard *mgd, midgard_server_config *scfg);

void mgd_set_apache_auth_env(request_rec *r, const char* user);

static GSList *database_connection = NULL;
static GSList *database_handle = NULL;

static char *default_rootfile = NULL;
static char *default_parser = NULL;

static int sitegroup_lib_mismatch = FALSE;

module midgard_module;

static const char *verify_user(midgard_request_config * rcfg)
{
	const char *username;
	const char *password;
	const char *cookie;
	const char *p;
	int reason = 0;

	/* Basic auth */
	if (mgd_get_basic_auth_pw(rcfg->req, &username, &password) == OK
	    && ((reason = mgd_auth(rcfg->mgd, username, password)) > 0)) {
		return username;
	}

	if (reason < 0)
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO,
			      rcfg->req, "Midgard: basic auth failed (%d)",
			      reason);

	/* cookie auth */
	if ((cookie = ap_table_get(rcfg->req->headers_in, "Cookie")) == NULL)
		return NULL;

	if ((cookie = strstr(cookie, "MidgardLogin=")) == NULL)
		return NULL;

	cookie += 13;		/* skip MidgardLogin= */

	if ((p = strchr(cookie, ';')) == NULL)
		return NULL;

	cookie = ap_pstrndup(rcfg->req->pool, cookie, p - cookie);
	cookie = ap_uudecode(rcfg->req->pool, cookie);
	if ((p = strchr(cookie, ':')) == NULL)
		return NULL;

	username = ap_pstrndup(rcfg->req->pool, cookie, p - cookie);
	password = ap_pstrdup(rcfg->req->pool, p + 1);

	if (mgd_auth(rcfg->mgd, username, password) > 0) {
		rcfg->auth.cookie = 1;
		return username;
	}

	return NULL;
}

static int mgd_find_pageblob(midgard_request_config * rcfg,
			     midgard_directory_config * dcfg, char *name)
{
	const char *content_type;
	const char *location;
	midgard_res *res;

	switch (rcfg->resource.type) {
		case MOD_MIDGARD_RESOURCE_PAGE:
			if (rcfg->resource.id == 0 || dcfg->blobdir == NULL)
				return DECLINED;
			break;
		case MOD_MIDGARD_RESOURCE_BLOB:
			return HTTP_NOT_FOUND;
		default:
			return HTTP_INTERNAL_SERVER_ERROR;
	}

	ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, rcfg->req,
		      "Midgard: locating blob=[%d]/%s", rcfg->resource.id,
		      name);

	res = mgd_ungrouped_select(rcfg->mgd, "id,location,mimetype",
				   "blobs",
				   "pid=$d AND ptable='page' AND name=$q",
				   NULL, rcfg->resource.id, name);

	if (!res || !mgd_fetch(res)) {
		if (res)
			mgd_release(res);
		return DECLINED;
	}

	location = mgd_colvalue(res, 1);
	content_type = mgd_colvalue(res, 2);

	if (!location || *location == '\0') {
		ap_log_rerror(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO,
			      rcfg->req,
			      "Midgard: blob %s ([%d]/%s) has empty location",
			      mgd_colvalue(res, 0), rcfg->resource.id, name);

		mgd_release(res);
		return HTTP_NOT_FOUND;
	}

	/* BlobDir must start with a '/' and if location looks like a
	   relative path we will not serve it.
	 */
	if (*(dcfg->blobdir) != '/' || strstr(location, "..")) {
		ap_log_rerror(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO,
			      rcfg->req,
			      "Midgard: BlobDir must start with '/' [%s]",
			      dcfg->blobdir);
		mgd_release(res);
		return HTTP_INTERNAL_SERVER_ERROR;
	}

	if (!content_type || *content_type == '\0')
		content_type = "application/binary";

	rcfg->path = ap_pstrcat(rcfg->pool, dcfg->blobdir, "/", location, NULL);
	rcfg->resource.content_type = ap_pstrdup(rcfg->pool, content_type);

	rcfg->resource.type = MOD_MIDGARD_RESOURCE_BLOB;
	rcfg->resource.id = atoi(mgd_colvalue(res, 0));

	mgd_release(res);

	ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, rcfg->req,
		      "Midgard: blob %d found, maps to %s (%s)",
		      rcfg->resource.id,
		      rcfg->path, rcfg->resource.content_type);

	return OK;
}

static void midgard_child_init(server_rec * s, pool * p)
{
	/* Initialize the default rootfile setting. */
	default_rootfile =
	   ap_server_root_relative(p, LIBEXEC_DIR "/midgard-root.php3");
}

static void midgard_disconnect_all(void *data)
{
GSList *list;
midgard_database_handle *dbh;
midgard_database_connection *dbc;

int connections = 0;

   if (database_handle) {
      for (list = g_slist_last(database_handle); list != NULL;
            list = g_slist_next(list)) {
         dbh = list->data;
         if (!dbh) continue;

         if (dbh->mgd != NULL) mgd_close(dbh->mgd);
         if (dbh->username) free(dbh->username);
         if (dbh->password) free(dbh->password);
         free(dbh);
         connections++;
      }

      g_slist_free(database_handle);
      database_handle = NULL;
   }

   ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, (server_rec*)data,
      "Midgard: closing %d database connections", connections);

   if (database_connection) {
      for (list = g_slist_last(database_connection); list != NULL;
            list = g_slist_next(list)) {
         dbc = list->data;
         if (!dbc) continue;

         if (dbc->name) free (dbc->name);
         free(dbc);
      }

      g_slist_free(database_connection);
      database_connection = NULL;
   }

   mgd_done();
}

static void midgard_child_exit(server_rec * s, pool * p)
{
   midgard_disconnect_all(s);
}

static void midgard_module_init(server_rec * s, pool * p)
{
#if MODULE_MAGIC_NUMBER >= 19980527
	const char *component;
	component = ap_pstrcat(p, "Midgard/", MIDGARD_LIB_VERSION, NULL);
	ap_add_version_component(component);
#endif

#if HAVE_MIDGARD_SITEGROUPS
   if (!midgard_lib_compiled_with_sitegroups()) {
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO,
			      s, "Midgard: midgard-lib does not support sitegroups");
      sitegroup_lib_mismatch = TRUE;
   }
#else
   if (midgard_lib_compiled_with_sitegroups()) {
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO,
			      s, "Midgard: mod_midgard does not support sitegroups");
      sitegroup_lib_mismatch = TRUE;
   }
#endif

	mgd_init();

	default_parser = ap_pstrdup(p, "latin1");

   /* EEH: Not the cleanest thing in the world, but a lot of modules use
      this. The cleanup function is called when the pool is deallocated,
      and this particular pool will be deallocated when the handler dies.
      Blargh.
   ap_register_cleanup(p, s, midgard_disconnect_all, ap_null_cleanup);
   */
}

/* String pointers are equal if both NULL or the strings are equal */
#define SAFECMP(a,b) ((!(a) && !(b)) || ((a) && (b) && !strcmp((a),(b))))

static midgard_database_connection *midgard_get_database_connection(
   pool * p,
   const char *database,
   const char *username,
   const char *password)
{
	midgard_database_connection *dbc;
   midgard_database_handle *dbh;
   GSList *list;

   dbh = NULL;

   for (list = g_slist_last(database_connection); list != NULL;
         list = g_slist_next(list)) {

      dbc = list->data;
      if (!dbc || ! dbc->handle) {
         fprintf(stderr, "Midgard: error in connection array\n");
         continue;
      }

      if (  ! SAFECMP(username, dbc->handle->username)
            || ! SAFECMP(password, dbc->handle->password) ) {
         continue;
      }

      if ( SAFECMP(database, dbc->name) ) {
         return dbc;
      }

      dbh = dbc->handle;
      break;
   }

   if (dbh == NULL) {
      dbh = malloc(sizeof(midgard_database_handle));
      dbh->username = strdup(username);
      dbh->password = strdup(password);
      dbh->mgd = NULL;
      dbh->current = NULL;

      if (database_handle == NULL) {
         database_handle = g_slist_alloc();
         database_handle->data = dbh;
      } else {
         database_handle = g_slist_append(database_handle, dbh);
      }
   }

   dbc = malloc(sizeof(midgard_database_connection));
   dbc->name = strdup(database);
   dbc->handle = dbh;

   if (database_connection == NULL) {
      database_connection = g_slist_alloc();
      database_connection->data = dbc;
   } else {
      database_connection = g_slist_append(database_connection, dbc);
   }

	return dbc;
}

static void *midgard_create_directory_config(pool * p, char *dir)
{
	midgard_directory_config *dcfg = (midgard_directory_config *)
	   ap_palloc(p, sizeof(midgard_directory_config));

	dcfg->engine.on = 0;
	dcfg->engine.set = 0;

	dcfg->extauth.on = 0;
	dcfg->extauth.set = 0;

	dcfg->rootfile = NULL;
	dcfg->templatefile = NULL;
	dcfg->blobdir = NULL;
	dcfg->parser = default_parser;

	return (void *) dcfg;
}

static void *midgard_merge_directory_config(pool * p, void *base, void *new)
{
	midgard_directory_config *dcfg_base = (midgard_directory_config *) base;
	midgard_directory_config *dcfg_new = (midgard_directory_config *) new;
	midgard_directory_config *dcfg = (midgard_directory_config *)
	   ap_palloc(p, sizeof(midgard_directory_config));

	dcfg->engine.on =
	   dcfg_new->engine.set ? dcfg_new->engine.on : dcfg_base->engine.on;
	dcfg->engine.set = dcfg_new->engine.set || dcfg_base->engine.set;

	dcfg->extauth.on =
	   dcfg_new->extauth.set ? dcfg_new->extauth.on : dcfg_base->extauth.on;
	dcfg->extauth.set = dcfg_new->extauth.set || dcfg_base->extauth.set;

	dcfg->rootfile = ap_pstrdup(p, dcfg_new->rootfile ?
				    dcfg_new->rootfile : dcfg_base->rootfile);

   if (dcfg_new->templatefile)
	   dcfg->templatefile = ap_pstrdup(p, dcfg_new->templatefile);
   else if (dcfg_base->templatefile)
	   dcfg->templatefile = ap_pstrdup(p, dcfg_base->templatefile);
   else
      dcfg->templatefile = NULL;

	dcfg->blobdir = ap_pstrdup(p, dcfg_new->blobdir ?
				   dcfg_new->blobdir : dcfg_base->blobdir);
	dcfg->parser = ap_pstrdup(p, dcfg_new->parser ?
				  dcfg_new->parser : dcfg_base->parser);

	return (void *) dcfg;
}

static void *midgard_create_server_config(pool * p, server_rec * s)
{
	midgard_server_config *scfg = (midgard_server_config *)
	   ap_palloc(p, sizeof(midgard_server_config));

	scfg->default_realm = "Midgard";

	scfg->database.page = NULL;
	scfg->database.auth = NULL;
	scfg->database.main = NULL;
	scfg->forcedroot.set = 0;

	return (void *) scfg;
}

static void *midgard_merge_server_config(pool * p, void *base, void *new)
{
	midgard_server_config *scfg_base = (midgard_server_config *) base;
	midgard_server_config *scfg_new = (midgard_server_config *) new;
	midgard_server_config *scfg = (midgard_server_config *)
	   ap_pcalloc(p, sizeof(midgard_server_config));

	/* Can only be used outside of VirtualHost directives */
	scfg->default_realm = ap_pstrdup(p, scfg_new->default_realm ?
				    scfg_new->default_realm : scfg_base->default_realm);

	/* Inherit the database to be used */
   scfg->database.page =
      scfg_new->database.page
         ? scfg_new->database.page
         : scfg_base->database.page;
   scfg->database.auth =
      scfg_new->database.auth
         ? scfg_new->database.auth
         : scfg_base->database.auth;
   scfg->database.main =
      scfg_new->database.main
         ? scfg_new->database.main
         : scfg_base->database.main;

	scfg->forcedroot.on =
	   scfg_new->forcedroot.set ? scfg_new->forcedroot.on : scfg_base->forcedroot.on;
	scfg->forcedroot.set = scfg_new->forcedroot.set || scfg_base->forcedroot.set;

	return (void *) scfg;
}

static const char *midgard_main_db_cmd(cmd_parms * cmd, void *dummy,
					char *database,
					char *username, char *password)
{
	midgard_server_config *scfg = (midgard_server_config *)
	   ap_get_module_config(cmd->server->module_config,
				&midgard_module);

	scfg->database.main = midgard_get_database_connection(cmd->pool,
         database, username, password);

	return NULL;
}

static const char *midgard_page_db_cmd(cmd_parms * cmd, void *dummy,
					char *database,
					char *username, char *password)
{
	midgard_server_config *scfg = (midgard_server_config *)
	   ap_get_module_config(cmd->server->module_config,
				&midgard_module);

	scfg->database.page = midgard_get_database_connection(cmd->pool,
         database, username, password);

	return NULL;
}

static const char *midgard_auth_db_cmd(cmd_parms * cmd, void *dummy,
					char *database,
					char *username, char *password)
{
	midgard_server_config *scfg = (midgard_server_config *)
	   ap_get_module_config(cmd->server->module_config,
				&midgard_module);

	scfg->database.auth = midgard_get_database_connection(cmd->pool,
         database, username, password);
   scfg->forcedroot.set = 1;
   scfg->forcedroot.on = 1;

	return NULL;
}

static const char *midgard_engine_cmd(cmd_parms * cmd, void *cfg, int flag)
{
	midgard_directory_config *dcfg = (midgard_directory_config *) cfg;

	dcfg->engine.on = flag;
	dcfg->engine.set = 1;

	return NULL;
}

static const char *midgard_auth_cmd(cmd_parms * cmd, void *cfg, int flag)
{
	midgard_directory_config *dcfg = (midgard_directory_config *) cfg;

	dcfg->extauth.on = flag;
	dcfg->extauth.set = 1;

	return NULL;
}

static const char *midgard_rootfile_cmd(cmd_parms * cmd, void *cfg,
					char *rootfile)
{
	midgard_directory_config *dcfg = (midgard_directory_config *) cfg;

	dcfg->rootfile = ap_server_root_relative(cmd->pool, rootfile);

	return NULL;
}

static const char *midgard_template_cmd(cmd_parms * cmd, void *cfg,
					char *template)
{
	midgard_directory_config *dcfg = (midgard_directory_config *) cfg;

	dcfg->templatefile = ap_pstrdup(cmd->pool, template);

	return NULL;
}

static const char *midgard_blobdir_cmd(cmd_parms * cmd, void *cfg,
				       char *blobdir)
{
	midgard_directory_config *dcfg = (midgard_directory_config *) cfg;

	dcfg->blobdir = ap_server_root_relative(cmd->pool, blobdir);

	return NULL;
}

#if HAVE_MIDGARD_SITEGROUPS
static const char *midgard_realm_cmd(cmd_parms * cmd, void *cfg, char *realm)
{
	midgard_server_config *scfg = (midgard_server_config *)
	   ap_get_module_config(cmd->server->module_config,
				&midgard_module);

//      if(realm && *realm)
	scfg->default_realm = ap_pstrdup(cmd->pool, realm);

	return NULL;
}
#else
static const char *midgard_realm_cmd(cmd_parms * cmd, void *cfg, char *realm)
{
	return NULL;
}
#endif

static const char *midgard_parser_cmd(cmd_parms * cmd, void *cfg, char *parser)
{
	midgard_directory_config *dcfg = (midgard_directory_config *) cfg;

	dcfg->parser = ap_pstrdup(cmd->pool, parser);

	return NULL;
}

static void *midgard_create_request_config(request_rec * r)
{
	midgard_request_config *rcfg = (midgard_request_config *)
	   ap_pcalloc(r->pool, sizeof(midgard_request_config));

	rcfg->req = r;
	rcfg->pool = r->pool;
	rcfg->mgd = NULL;

	rcfg->host = 0;
	rcfg->style = 0;
	rcfg->auth.required = 0;
	rcfg->auth.passed = 0;
	rcfg->auth.cookie = 0;
	rcfg->author = 0;

	rcfg->prelen = 0;
	rcfg->self_len = 0;

	rcfg->resource.id = 0;
	rcfg->resource.status = DECLINED;
	rcfg->resource.type = MOD_MIDGARD_RESOURCE_PAGE;
	rcfg->resource.active_page = 0;

	rcfg->resource.content_type = NULL;
	rcfg->path = NULL;

	rcfg->argv = ap_make_array(r->pool, 8, sizeof(const char *));

	rcfg->elements = ap_make_table(r->pool, 8);

   rcfg->database.current = NULL;
   rcfg->database.page = NULL;
   rcfg->database.main = NULL;

	return (void *) rcfg;
}

struct _ftstatus
{
	int status;
	int host;
	int style;
	int sitegroup;

	int auth_required;
	int stopped;
	int skip_host_style;
	char *filename;

	midgard_request_config *rcfg;
};

static void midgard_get_host(midgard_request_config * rcfg,
			     struct _ftstatus *ftstatus)
{
	midgard_res *res;
	int prelen;

	ftstatus->status = DECLINED;

	res = mgd_ungrouped_select(rcfg->mgd,
				   "id,style,root,info&1,Length(prefix)"
#if HAVE_MIDGARD_SITEGROUPS
				   ",sitegroup"
#endif
				   , "host",
				   "name=$q AND port IN (0,$d) AND"
				   " prefix=Left($q,Length(prefix)) AND"
				   " IF($d>=Length(prefix),MID($q,1+Length(prefix),1) IN ('','/'),0)=1"
				   " AND online=1",
				   "prefix DESC, port DESC",
				   ap_get_server_name(rcfg->req),
				   ap_get_server_port(rcfg->req),
				   rcfg->req->uri, strlen(rcfg->req->uri),
				   rcfg->req->uri);
	if (!res)
		return;

	if (mgd_fetch(res)) {
		rcfg->host = atoi(mgd_colvalue(res, 0));
		rcfg->style = atoi(mgd_colvalue(res, 1));
		rcfg->resource.type = MOD_MIDGARD_RESOURCE_PAGE;
		rcfg->resource.status = OK;
		rcfg->resource.id = atoi(mgd_colvalue(res, 2));
		rcfg->auth.required = atoi(mgd_colvalue(res, 3));
		prelen = atoi(mgd_colvalue(res, 4));
		rcfg->prelen = prelen;
#if HAVE_MIDGARD_SITEGROUPS
		ftstatus->sitegroup = atoi(mgd_colvalue(res, 5));
		mgd_set_sitegroup(rcfg->mgd, ftstatus->sitegroup);
#endif

		if (prelen == 0) {
			ftstatus->status = OK;
			ftstatus->host = rcfg->host;
			ftstatus->style = rcfg->style;
			ftstatus->auth_required = rcfg->auth.required;
		}
		else
			while (prelen != 0 && mgd_fetch(res)) {

				if ((prelen = atoi(mgd_colvalue(res, 4))) != 0)
					continue;

				ftstatus->status = OK;
				ftstatus->host = atoi(mgd_colvalue(res, 0));
				ftstatus->style = atoi(mgd_colvalue(res, 1));
				ftstatus->auth_required =
				   atoi(mgd_colvalue(res, 3));
#if HAVE_MIDGARD_SITEGROUPS
				ftstatus->sitegroup =
				   atoi(mgd_colvalue(res, 5));
#endif
			}
	}

	mgd_release(res);
}

static void midgard_get_elements(midgard_request_config * rcfg, int style)
{
	midgard_res *res;

	res =
	   mgd_ungrouped_select
	   (rcfg->mgd, "name,value", "element", "style=$d", NULL, style);
	if (!res)
		return;

	while (mgd_fetch(res))
		if (!ap_table_get(rcfg->elements, mgd_colvalue(res, 0)))
			ap_table_set(rcfg->elements, mgd_colvalue(res, 0),
				     mgd_colvalue(res, 1));

	mgd_release(res);
}

static void midgard_style_walk(midgard_request_config * rcfg)
{
	midgard_res *res;
	int style;

	/* EEH: this could select a custom mimetype for a specific style from a
	 * table. I'm not entirely sure this is done right, what with
	 * inheretance, and it isn't enabled now.
	 */
	res = mgd_ungrouped_record(rcfg->mgd, "type", "style", rcfg->style);
	if (res && mgd_fetch(res))
		rcfg->resource.content_type
		   = ap_pstrdup(rcfg->pool, mgd_colvalue(res, 0));
	if (res)
		mgd_release(res);

	for (style = rcfg->style; style;
	     style = mgd_idfield(rcfg->mgd, "up", "style", style))
		midgard_get_elements(rcfg, style);
}

static void midgard_get_pelements(midgard_request_config * rcfg, int inherit)
{
	midgard_res *res;

	res =
	   mgd_ungrouped_select
	   (rcfg->mgd, "name,value", "pageelement",
	    "page=$d AND info&1=$d", NULL, rcfg->resource.id, inherit);
	if (!res)
		return;

	while (mgd_fetch(res))
		ap_table_set(rcfg->elements, mgd_colvalue(res, 0),
			     mgd_colvalue(res, 1));

	mgd_release(res);
}

static void midgard_find_pathelement(midgard_directory_config * dcfg,
				     midgard_request_config * rcfg, char *path)
{
	midgard_res *res = NULL;
	int tmp;
#if HAVE_MIDGARD_PAGELINKS
	int pagelink_ok=0;
#endif

	if (rcfg->resource.type != MOD_MIDGARD_RESOURCE_PAGE) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO,
			      rcfg->req,
			      "Midgard: %s has non-page parent %d:%d", path,
			      rcfg->resource.type, rcfg->resource.id);
		rcfg->resource.status = DECLINED;
		return;
	}

#if HAVE_MIDGARD_PAGELINKS
	/* try link first */
	// If the user isn't yet authenticated, it will go to a 
	// default page (if grp=0 exists in pagelink) or the pagelink isn't
	// available (404 or active page or blob).
	// if the user is already authenticated, it will go to the linked
	// page depending on its group membership (or the default page if
	// it's not a member of any defined pagelink grp)
	// Note: the pagelink does not restart walking pages from the root,
	// so this may cause surprises if the style and/or page elements are
	// not the ones we want.
	if (midgard_authenticate(rcfg) == OK) {
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO,
		      rcfg->req,
		      "Midgard: resolving path element: user #%d searches link %s",
		     	 mgd_user(rcfg->mgd), path);

		res = mgd_ungrouped_select
		   (rcfg->mgd,
		    "distinct page.id,page.style,page.info&1,page.info&2",
		    "pagelink,page,member",
		    "pagelink.up=$d AND pagelink.target=page.id "
			    		   "AND pagelink.name=$q "
						   "AND (pagelink.grp=0 OR (pagelink.grp=member.gid "
						   "AND member.uid=$d)) "
		    ,"pagelink.grp DESC", 
		    rcfg->resource.id, path, mgd_user(rcfg->mgd));
		if (res) {
			pagelink_ok = mgd_fetch(res);
		}
	}
	if(!pagelink_ok) {
#endif

	/* find page */

	res = mgd_ungrouped_select(rcfg->mgd, "id,style,info&1,info&2",
				   "page",
				   "up=$d AND name=$q", NULL,
				   rcfg->resource.id, path);

#if HAVE_MIDGARD_PAGELINKS
	} // !pagelink_ok
#endif

	if (res
#if HAVE_MIDGARD_PAGELINKS
			&& (pagelink_ok || mgd_fetch(res)) 
#else
	       	&& mgd_fetch(res)
#endif
			) {
		/* not set because it's the default:
		   rcfg->resource.type = MOD_MIDGARD_RESOURCE_PAGE;
		 */
		rcfg->resource.status = OK;
		rcfg->resource.id = atoi(mgd_colvalue(res, 0));

		if ((tmp = atoi(mgd_colvalue(res, 1))) != 0)
			rcfg->style = tmp;

		if ((tmp = atoi(mgd_colvalue(res, 2))) != 0)
			rcfg->auth.required = tmp;

		rcfg->resource.active_page = atoi(mgd_colvalue(res, 3));

		mgd_release(res);

		/* Get inherited page elements */
		midgard_get_pelements(rcfg, 1);

		return;
	}

	/* no page or link, try blob */
	if (res != NULL)
		mgd_release(res);

	rcfg->resource.status = mgd_find_pageblob(rcfg, dcfg, path);
}

static void midgard_page_walk(midgard_request_config * rcfg,
			      midgard_directory_config * dcfg)
{
	midgard_res *res;
	char *uri, *tok, *part, **arg;
	int tmp, addslash;

   rcfg->self_len = rcfg->prelen;

	tok = strrchr(rcfg->req->uri + rcfg->prelen, '/');
	part = tok ? strchr(tok + 1, '.') : NULL;
	addslash = !tok || (*(tok + 1) != '\0' && !part);

	part = strrchr(rcfg->req->uri + rcfg->prelen, '.');
	if (part && strcmp(part, ".html") == 0) {
		uri = ap_pstrndup(rcfg->pool, rcfg->req->uri + rcfg->prelen,
				  part - rcfg->req->uri - rcfg->prelen);
	}
	else {
		uri = ap_pstrdup(rcfg->pool, rcfg->req->uri + rcfg->prelen);
	}

	/* Get root page information */
	rcfg->resource.active_page = mgd_idfield(rcfg->mgd, "info&2",
						 "page", rcfg->resource.id);

	/* Get inherited page elements of the root page */
	midgard_get_pelements(rcfg, 1);

	/* Traverse the page structure */
	for (tok = strsep(&uri, "/"); tok; tok = strsep(&uri, "/")) {
		if (!*tok) { rcfg->self_len += 1; continue; }

		midgard_find_pathelement(dcfg, rcfg, tok);

		if (rcfg->resource.status != OK) {
			if (rcfg->resource.type != MOD_MIDGARD_RESOURCE_PAGE
			    || !rcfg->resource.active_page) {
				return;
			}

			/* restore status for active page */
			rcfg->resource.status = OK;
			break;
		}

      rcfg->self_len += strlen(tok) + 1;
	}

   rcfg->self_len--;

	if (rcfg->resource.type == MOD_MIDGARD_RESOURCE_BLOB)
		return;
	if (rcfg->resource.type == MOD_MIDGARD_RESOURCE_PAGE) {
		/* Scan path arguments for active pages */
		if (rcfg->resource.active_page) {
			while (tok) {
            addslash = 0;
            if (*tok) {
				   arg = (char **) ap_push_array(rcfg->argv);
				   *arg = ap_pstrdup(rcfg->pool, tok);
            }
				tok = strsep(&uri, "/");
			}
         if (!addslash) rcfg->self_len++;
		}
		
		if (addslash) {
			rcfg->resource.type =
			   MOD_MIDGARD_RESOURCE_PAGE_REDIRECT;
			return;
		}
	}

	/* Get rest of the page elements (not inherited) of the page */
	midgard_get_pelements(rcfg, 0);

	/* Get contents of the page */
	res =
	   mgd_ungrouped_record(rcfg->mgd, "title,content,author,style,info&1",
				"page", rcfg->resource.id);
	if (!res || !mgd_fetch(res)) {
		if (res)
			mgd_release(res);

		rcfg->resource.status = DECLINED;
		return;
	}

	ap_table_set(rcfg->elements, "title", mgd_colvalue(res, 0));
	ap_table_set(rcfg->elements, "content", mgd_colvalue(res, 1));

	rcfg->author = atoi(mgd_colvalue(res, 2));

	if ((tmp = atoi(mgd_colvalue(res, 3))) != 0)
		rcfg->style = tmp;
	if ((tmp = atoi(mgd_colvalue(res, 4))) != 0)
		rcfg->auth.required = tmp;

	mgd_release(res);

	res = mgd_ungrouped_record(rcfg->mgd, "CONCAT(firstname,' ',lastname)",
				   "person", rcfg->author);
	if (!res || !mgd_fetch(res)) {
		if (res)
			mgd_release(res);

		/* EEH Why? */
		rcfg->resource.status = DECLINED;
		return;
	}

	ap_table_set(rcfg->elements, "author", mgd_colvalue(res, 0));
	mgd_release(res);
}

int midgard_authenticate(midgard_request_config * rcfg)
{
	const char *username = NULL;

	if (rcfg->auth.passed)
		return OK;

	if (!rcfg->auth.required)
		return OK;

	if ((username = verify_user(rcfg)) == NULL) {
		mgd_note_basic_auth_failure(rcfg->req, rcfg->mgd, rcfg->scfg);
		return AUTH_REQUIRED;
	}

	if (!rcfg->auth.cookie) {
		mgd_set_apache_auth_env(rcfg->req, username);
	}

	rcfg->auth.passed = 1;
   if (rcfg->scfg->forcedroot.set && rcfg->scfg->forcedroot.on) {
#if HAVE_MIDGARD_SITEGROUPS
      mgd_force_root(rcfg->mgd);
#else
      mgd_force_admin(rcfg->mgd);
#endif
   }

	return OK;
}

int mod_mgd_ft_filename(char *filename, void *client_data)
{
	((struct _ftstatus *) client_data)->filename = filename;
	return 1;
}

int mod_mgd_ft_signal(int ft_sig, void *client_data)
{
	struct _ftstatus *ftstatus = (struct _ftstatus *) client_data;

	switch (ft_sig) {
		case MGD_FT_SIGNAL_AUTH:
			ftstatus->auth_required = 1;
			break;

		case MGD_FT_SIGNAL_STOP:
			ftstatus->stopped = 1;
			break;

		case MGD_FT_SIGNAL_SKIP_HOST_STYLE:
			ftstatus->skip_host_style = 1;
			break;

		case MGD_FT_SIGNAL_PAGE_IN_STYLE:
			ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO,
				      ftstatus->rcfg->req,
				      "Midgard: page record in %s",
				      ftstatus->filename);
			break;

		default:
			ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO,
				      ftstatus->rcfg->req,
				      "Midgard: unexpected filetemplate signal %d",
				      ft_sig);
			return 0;
	}

	return 1;
}

int mod_mgd_ft_element(char *name, char *value, void *client_data)
{
	struct _ftstatus *ftstatus = (struct _ftstatus *) client_data;

	if (!ap_table_get(ftstatus->rcfg->elements, name))
		ap_table_set(ftstatus->rcfg->elements, name, value);

	return 1;
}

void try_filetemplates(request_rec * r,
            midgard_request_config * rcfg,
            midgard_directory_config * dcfg,
            struct _ftstatus *ftstatus)
{
	int status;
	request_rec *subreq;
   char *path_info;
   char *tok;
   char **arg;

	/* host not found */
	if (ftstatus->status != OK) {
      /* EEH: no host match, no connected templates */
		return;
	}

	/* determine translated filename + handler by subrequest */
	if ((subreq = ap_sub_req_lookup_uri(r->uri, r)) == NULL) {
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		      "Midgard: filetemplate request subrequest failed");
		rcfg->resource.status = DECLINED;
		return;
	}

	ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
	      "Midgard: filetemplate request handler %s", subreq->handler);
	rcfg->resource.status = DECLINED;
	if (!subreq->filename || !subreq->content_type
	    || strcmp(subreq->content_type, MIDGARD_HANDLER) != 0) {
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		      "Midgard: filetemplate request for %s found no filename (%s) or mimetype (%s)",
               r->uri, subreq->filename ? subreq->filename : "<NULL>",
               subreq->content_type ? subreq->content_type : "<NULL>");
		ap_destroy_sub_req(subreq);
		rcfg->resource.status = DECLINED;
		return;
	}

   ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
      "Midgard: filetemplate request for %s (%s) maps to file %s",
         r->uri, subreq->content_type, subreq->filename);

	ftstatus->filename = NULL;
	ftstatus->rcfg = rcfg;
	ftstatus->auth_required = 0;
	ftstatus->skip_host_style = 0;
	ftstatus->stopped = 0;

	status = midgard_traverse_filetemplate(subreq->filename,
                     dcfg->templatefile,
                     mod_mgd_ft_signal,
                     mod_mgd_ft_element,
                     mod_mgd_ft_filename,
                     NULL, ftstatus);

	ap_destroy_sub_req(subreq);

	switch (status) {
		case MGD_FT_ERR_NONE:
			rcfg->host = ftstatus->host;
			rcfg->style =
			   ftstatus->skip_host_style ? 0 : ftstatus->style;
			rcfg->resource.type = MOD_MIDGARD_RESOURCE_FILETEMPLATE;
			rcfg->resource.status = OK;
			rcfg->resource.id = 0;
			rcfg->auth.required |= ftstatus->auth_required;
			rcfg->prelen = 0;
			rcfg->self_len = strlen(r->uri);
	      rcfg->path = ap_pstrdup(rcfg->pool, subreq->filename);
#if HAVE_MIDGARD_SITEGROUPS
			mgd_set_sitegroup(rcfg->mgd, ftstatus->sitegroup);
#endif

	      tok = strrchr(subreq->path_info, '.');
	      if (tok && strcmp(tok, ".html") == 0) {
		      path_info = ap_pstrndup(rcfg->pool, subreq->path_info,
               tok - subreq->path_info);
	      }
	      else {
		      path_info = ap_pstrdup(rcfg->pool, subreq->path_info);
	      }

         while ((tok = strsep(&path_info, "/")) != NULL) {
            if (*tok) {
				   arg = (char **) ap_push_array(rcfg->argv);
				   *arg = ap_pstrdup(rcfg->pool, tok);
            }
			}

			if (!ftstatus->stopped) {
				ap_log_rerror(APLOG_MARK,
					      APLOG_WARNING | APLOG_NOERRNO, r,
					      "Midgard: no filetemplate stop, parsed all the way to the root");
			}
			break;

		case MGD_FT_ERR_NWF:
			ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
				      "Midgard: %s not well-formed (%s requested)",
				      ftstatus->filename ? ftstatus->filename : "<NULL>", subreq->filename);
			break;

		case MGD_FT_ERR_ACCESS:
			ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
				      "Midgard: access to %s denied",
				      subreq->filename);
			break;

		case MGD_FT_ERR_PATH:
			ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
				      "Midgard: invalid path %s",
				      subreq->filename);
			break;

		case MGD_FT_ERR_ABORT:
			/* the aborting routine will do the logging */
			break;

		default:
			ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
				      "Midgard: unknown filetemplate status %d for %s",
				      status, subreq->filename);
	}
}

static int midgard_translate_handler(request_rec * r)
{
	midgard_server_config *scfg;
	midgard_directory_config *dcfg;
	midgard_request_config *rcfg;
	const char *uri, *parser;
	struct _ftstatus ftstatus;

	/* we don't do subrequests */
	if (r->main != NULL)
		return DECLINED;

   /* midgard-lib was recompiled with a different sitegroup setting, bail */
   if (sitegroup_lib_mismatch)
      return DECLINED;

	/* Check whether MidgardEngine is on */
	dcfg = (midgard_directory_config *)
	   ap_get_module_config(r->per_dir_config, &midgard_module);
	if (!dcfg || !dcfg->engine.on)
		return DECLINED;

	ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		      "midgard_translate_handler(%s, http://%s:%d%s)",
		      dcfg->parser ?
		      dcfg->parser : default_parser,
		      ap_get_server_name(r), ap_get_server_port(r), r->uri);

	/* Check that the Midgard database is available */
	scfg = (midgard_server_config *)
	   ap_get_module_config(r->server->module_config, &midgard_module);

	if (!scfg || !scfg->database.main) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
			      "Midgard: database not defined for %s",
			      ap_get_server_name(r));
		return DECLINED;
	}

	/* Create the request config record */
	rcfg = (midgard_request_config *) midgard_create_request_config(r);
	rcfg->scfg = scfg;

	if (!mgd_select_database(r, scfg->database.main, rcfg, TRUE)) return DECLINED;

	if (scfg->database.main->handle == NULL || scfg->database.main->handle->mgd == NULL) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
			"Midgard: error opening database connection for %s to %s as %s",
				ap_get_server_name(r),
				scfg->database.main->name, 
				scfg->database.main->handle->username
			);
		return DECLINED;
	}

	mgd_clear(rcfg->mgd);
	parser = dcfg->parser ? dcfg->parser : default_parser;
	mgd_select_parser(rcfg->mgd, parser);

   rcfg->database.page = scfg->database.page;
   rcfg->database.main = scfg->database.main;

   if (scfg->database.page != NULL) {
      if (!mgd_select_database(r, scfg->database.page, rcfg, TRUE)) return DECLINED;
   }

	/* Find host record */
	midgard_get_host(rcfg, &ftstatus);
	if (!rcfg->host) {
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
			      "Midgard: host record for %s:%d not found",
			      ap_get_server_name(r), ap_get_server_port(r));
		return DECLINED;
	}

	/* Directory walk */
	midgard_page_walk(rcfg, dcfg);

	if (dcfg->templatefile != NULL && rcfg->resource.status != OK)
		try_filetemplates(r, rcfg, dcfg, &ftstatus);

	if (rcfg->resource.status != OK) {
		return rcfg->resource.status;
	}

   mgd_set_blobdir(rcfg->mgd, dcfg->blobdir);

   if (scfg->database.auth != NULL) {
      if (!mgd_select_database(r, scfg->database.auth, rcfg, TRUE)) return DECLINED;
   }
   else {
      if (!mgd_select_database(r, scfg->database.main, rcfg, FALSE)) return DECLINED;
   }

	if (midgard_authenticate(rcfg) == AUTH_REQUIRED)
		return AUTH_REQUIRED;

	switch (rcfg->resource.type) {
		case MOD_MIDGARD_RESOURCE_PAGE_REDIRECT:
			uri = r->args
			   ? ap_pstrcat(r->pool,
					ap_escape_uri(r->pool, r->uri), "/",
					"?", r->args, NULL)
			   : ap_pstrcat(r->pool,
					ap_escape_uri(r->pool, r->uri), "/",
					NULL);

			ap_table_set(r->headers_out, "Location",
				     ap_construct_url(r->pool, uri, r));
			return HTTP_MOVED_PERMANENTLY;

		case MOD_MIDGARD_RESOURCE_PAGE:
		case MOD_MIDGARD_RESOURCE_FILETEMPLATE:
			/* Style walk */
         if (scfg->database.page != NULL) {
            if (!mgd_select_database(r, scfg->database.page, rcfg, FALSE)) return DECLINED;
         }

			midgard_style_walk(rcfg);
			ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO,
				      r,
				      "Midgard: using style #%d (%d elements)",
				      rcfg->style,
				      ap_table_elts(rcfg->elements)->nelts);

			/* Point Midgard/PHP to the root file */
			r->filename =
			   dcfg->rootfile ? dcfg->rootfile : default_rootfile;
         if (!mgd_select_database(r, scfg->database.main, rcfg, FALSE)) return DECLINED;
			break;

		case MOD_MIDGARD_RESOURCE_BLOB:
			r->filename = (char *) rcfg->path;
			break;

		default:
			ap_log_rerror(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO,
				      r, "Midgard: Unknown resource type %d",
				      rcfg->resource.type);
			return HTTP_INTERNAL_SERVER_ERROR;
	}

	/* Store the gathered information for Midgard/PHP */
	ap_set_module_config(r->request_config, &midgard_module, (void *) rcfg);
	return OK;
}

static int midgard_type_checker(request_rec * r)
{
	midgard_request_config *rcfg = (midgard_request_config *)
	   ap_get_module_config(r->request_config, &midgard_module);

	if (!rcfg) return DECLINED;

	if (rcfg->resource.type == MOD_MIDGARD_RESOURCE_BLOB) {
	   r->content_type = rcfg->resource.content_type;
      return OK;
   }

   return DECLINED;
}

static int midgard_host_auth(request_rec * r)
{
	struct _ftstatus ftstatus;
	midgard_request_config rcfg;
	midgard_server_config *scfg;
	midgard_directory_config *dcfg;
	const char *parser;

	/* Check whether MidgardEngine is on */
	dcfg = (midgard_directory_config *)
	   ap_get_module_config(r->per_dir_config, &midgard_module);
	if (!dcfg || !dcfg->extauth.on)
		return DECLINED;

	/* Check that the Midgard database is available */
	scfg = (midgard_server_config *)
	   ap_get_module_config(r->server->module_config, &midgard_module);

	if (!scfg || (!scfg->database.auth && !scfg->database.main) ||
         (!scfg->database.auth->handle && !scfg->database.main->handle)) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
			      "Midgard: (authen) database not defined for %s",
			      ap_get_server_name(r));
		return DECLINED;
	}

   if (scfg->database.auth) mgd_select_database(r, scfg->database.auth, &rcfg, TRUE);
   else mgd_select_database(r, scfg->database.main, &rcfg, TRUE);

	if (!rcfg.mgd)
		return DECLINED;

	rcfg.scfg = scfg;
	mgd_clear(rcfg.mgd);
	parser = dcfg->parser ? dcfg->parser : default_parser;
	mgd_select_parser(rcfg.mgd, parser);

	rcfg.req = r;

	midgard_get_host(&rcfg, &ftstatus);
	if (ftstatus.status != OK || !ftstatus.host) {
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
			      "Midgard: host record for %s:%d not found",
			      ap_get_server_name(r), ap_get_server_port(r));
		mgd_close(rcfg.mgd);
		return DECLINED;
	}

	rcfg.host = ftstatus.host;
	rcfg.prelen = 0;
	rcfg.auth.required = 1;
	rcfg.auth.passed = 0;
#if HAVE_MIDGARD_SITEGROUPS
	mgd_set_sitegroup(rcfg.mgd, ftstatus.sitegroup);
#endif
   mgd_select_database(r, scfg->database.main, &rcfg, FALSE);
	if (midgard_authenticate(&rcfg) == AUTH_REQUIRED) {
		mgd_close(rcfg.mgd);
		return AUTH_REQUIRED;
	}

	if (rcfg.auth.cookie) {
		return DECLINED;
	}

	mgd_close(rcfg.mgd);
	return OK;
}


#define MGD_CONF (RSRC_CONF | ACCESS_CONF)

static const command_rec midgard_cmds[] = {
	{"MidgardDatabase", midgard_main_db_cmd, NULL, MGD_CONF, TAKE3,
	 "Name of the Midgard database, username and password from which to fetch content"},
	{"MidgardPageDatabase", midgard_page_db_cmd, NULL, MGD_CONF, TAKE3,
	 "Name of the Midgard database, username and password from which to fetch pages"},
	{"MidgardAuthDatabase", midgard_auth_db_cmd, NULL, MGD_CONF, TAKE3,
	 "Name of the Midgard database, username and password from which to authenticate"},
	{"MidgardEngine", midgard_engine_cmd, NULL, MGD_CONF, FLAG,
	 "On or Off to enable or disable (default) the Midgard engine"},
	{"MidgardRootfile", midgard_rootfile_cmd, NULL, MGD_CONF, TAKE1,
	 "Location of the Midgard root file"},
	{"MidgardParser", midgard_parser_cmd, NULL, MGD_CONF, TAKE1,
	 "Select Midgard text parser (latin1|russian)"},
	{"MidgardBlobDir", midgard_blobdir_cmd, NULL, MGD_CONF, TAKE1,
	 "Location of the Midgard BLOBs"},
	{"MidgardDefaultRealm", midgard_realm_cmd, NULL, MGD_CONF, TAKE1,
	 "Default Midgard Basic Auth Realm"},
	{"MidgardAuth", midgard_auth_cmd, NULL, MGD_CONF, FLAG,
	 "On or Off to enable or disable (default) Midgard Auth for non-midgard requests"},
	{"MidgardTemplate", midgard_template_cmd, NULL, MGD_CONF, TAKE1,
	 "Name of the Midgard filetemplate style element files"},
	{NULL}
};

module midgard_module = {
	STANDARD_MODULE_STUFF,
	midgard_module_init,
	midgard_create_directory_config,
	midgard_merge_directory_config,
	midgard_create_server_config,
	midgard_merge_server_config,
	midgard_cmds,
	NULL,			/* [7] list of handlers */
	midgard_translate_handler,	/* [2] */
	midgard_host_auth,	/* [5] */
	NULL,			/* [6] */
	NULL,			/* [4] */
	midgard_type_checker,	/* [7] */
	NULL,			/* [8] fixups */
	NULL,			/* [10] logger */
	NULL,			/* [3] header parser */
	midgard_child_init
#if MODULE_MAGIC_NUMBER >= 19970728
	, midgard_child_exit	/* process exit/cleanup */
#endif
#if MODULE_MAGIC_NUMBER >= 19970902
	   , NULL		/* [1] post read_request handling */
#endif
};

/* ========== UTILITY ROUTINES ================ */

int mgd_get_basic_auth_pw(request_rec * r, const char **usr, const char **pw)
{
	const char *auth_line = ap_table_get(r->headers_in,
					     r->proxyreq
#if APACHE_RELEASE >= 10310000
					     == STD_PROXY
#endif
					     ? "Proxy-Authorization" :
					     "Authorization");

	const char *t;

	if (!auth_line) {
		return AUTH_REQUIRED;
	}

	if (strcasecmp(ap_getword(r->pool, &auth_line, ' '), "Basic")) {
		/* Client tried to authenticate using wrong auth scheme */
		ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, r,
			      "client used wrong authentication scheme: %s",
			      r->uri);
		return AUTH_REQUIRED;
	}

	/* CHARSET_EBCDIC Issue's here ?!? Compare with 32/9 instead
	 * as we are operating on an octed stream ?
	 */
	while (*auth_line == ' ' || *auth_line == '\t')
		auth_line++;

#if APACHE_RELEASE >= 10309000
	t = ap_pbase64decode(r->pool, auth_line);
#else
	t = ap_uudecode(r->pool, auth_line);
#endif

	/* Note that this allocation has to be made from r->connection->pool
	 * because it has the lifetime of the connection.  The other allocations
	 * are temporary and can be tossed away any time.
	 */
	*usr = ap_getword_nulls(r->connection->pool, &t, ':');
	*pw = t;

	return OK;
}

void mgd_set_apache_auth_env(request_rec * r, const char *user)
{
	r->connection->user = (char *) user;
	r->connection->ap_auth_type = "Basic";
}

void mgd_note_basic_auth_failure(request_rec * r, midgard * mgd,
				 midgard_server_config * scfg)
{
	char *realm = "Basic realm = \"Midgard\"";

#if HAVE_MIDGARD_SITEGROUPS
	midgard_res *res;
	int sitegroup = mgd_sitegroup(mgd);

	if (sitegroup == 0)
		realm =
		   ap_pstrcat(r->pool, "Basic realm = \"", scfg->default_realm,
			      " (SG0)\"", NULL);
	else {
		res =
		   mgd_ungrouped_select(mgd, "name,realm", "sitegroup", "id=$d",
					NULL, sitegroup);

		if (!res || !mgd_fetch(res)) {
			if (res)
				mgd_release(res);
			realm =
			   ap_pstrcat(r->pool, "Basic realm = \"",
				      scfg->default_realm, NULL);
		   ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_WARNING, r,
			      "Midgard: failed to fetch sitegroup name and realm for "
               "sitegroup %d", sitegroup);
		}
		else {
			realm =
			   ap_pstrcat(r->pool, "Basic realm = \"",
				      mgd_colvalue(res, 1), " (",
				      mgd_colvalue(res, 0), ")\"", NULL);
		}
	}
#else
   realm = ap_pstrcat(r->pool, "Basic realm = \"",
      scfg->default_realm ? scfg->default_realm : "Midgard", NULL);
#endif

	ap_table_setn(r->err_headers_out, r->proxyreq
#if APACHE_RELEASE >= 10310000
		      == STD_PROXY
#endif
		      ? "Proxy-Authenticate" : "WWW-Authenticate", realm);
}
